/*
 * Copyright (c) 2017, Red Hat Inc. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package org.openjdk.gcbench.fragger;

import org.openjdk.gcbench.tests.UnderPressureTest;
import org.openjdk.gcbench.util.AllocProfileSupport;
import org.openjdk.gcbench.util.ratelimit.MultiTokenBucket;
import org.openjdk.jmh.annotations.*;

import java.util.concurrent.ThreadLocalRandom;
import java.util.concurrent.TimeUnit;

@Warmup(iterations = 5, time = 1, timeUnit = TimeUnit.SECONDS)
@Measurement(iterations = 5, time = 1, timeUnit = TimeUnit.SECONDS)
@Fork(1)
@BenchmarkMode(Mode.Throughput)
@OutputTimeUnit(TimeUnit.SECONDS)
@Threads(Threads.MAX)
@State(Scope.Benchmark)
public class TreeFragger extends UnderPressureTest {

    @Param("16")
    int payloadSize;

    @Param({"1000"})
    long ldsMB;

    @Param({"10"})
    int rate;

    Node root;

    int count;

    MultiTokenBucket bucket;

    Object o0, o16;

    private int getSizePerCount() {
        if (AllocProfileSupport.isAvailable()) {
            long a1 = AllocProfileSupport.getAllocatedBytes();
            o0 = create(0);
            long a2 = AllocProfileSupport.getAllocatedBytes();
            o16 = create(16);
            long a3 = AllocProfileSupport.getAllocatedBytes();

            long sizeO0 = a2 - a1;
            long sizeO16 = a3 - a2;

            return (int) ((sizeO16 - sizeO0) / 16);
        } else {
            // best guess
            return 24 + // Node
                   16 + payloadSize; // payload
        }
    }

    @Setup
    public void setup() {
        bucket = new MultiTokenBucket(rate);
        count = (int)Math.max(1, ldsMB * 1024 * 1024 / getSizePerCount());
        root = create(count);
    }

    private Node create(int count) {
        Node root = new Node(new Object());

        for (int addr = 0; addr < count; addr++) {
            Node cur = root;
            for (int m = 31 - Integer.numberOfLeadingZeros(addr); m >= 0; m--) {
                if ((addr & (1 << m)) != 0) {
                    if (cur.left == null) {
                        cur.left = new Node(uninitAlloc(payloadSize));
                    }
                    cur = cur.left;
                } else {
                    if (cur.right == null) {
                        cur.right = new Node(uninitAlloc(payloadSize));
                    }
                    cur = cur.right;
                }
            }
        }

        return root;
    }

    @Benchmark
    public void test() {
        bucket.limit();
        doStore(ThreadLocalRandom.current().nextInt(count), uninitAlloc(payloadSize));
    }

    private void doStore(int addr, Object obj) {
        Node cur = root;
        for (int m = 31 - Integer.numberOfLeadingZeros(addr); m >= 0; m--) {
            if ((addr & (1 << m)) != 0) {
                cur = cur.left;
            } else {
                cur = cur.right;
            }
        }
        cur.payload = obj;
    }

    static class Node {
        Node left;
        Node right;
        Object payload;

        public Node(Object payload) {
            this.payload = payload;
        }
    }

}
